//MH  1/5/2006  \Mart120\Include\influenc.h
#pragma once

#include "RObject.h"
#include "Vector.h"
#include "Matrix34.h"
#include "Macros.h"
#include "RGBFloat.h"
#include "Samplers.h"

class Ray;

class QuadObject : public RObject {
public:
    float m_smin{0.f}, m_smax{0.f}, m_tmin{0.f}, m_tmax{0.f}; //16 bytes

    void *operator new(const size_t size) {
        return AllocSubPatch(size);
    }

    void operator delete(void *) {} //freed by dumping the whole pool

    QuadObject() : RObject(QUADRANGLE) {}
    explicit QuadObject(const UCHAR nodetype) : RObject(nodetype) {}
    QuadObject(const float m1, const float m2, const float t1, const float t2) : RObject(QUADRANGLE), m_smin(m1), m_smax(m2), m_tmin(t1), m_tmax(t2) {}

    bool IsPatch() final {
        return true;
    }

    UCHAR GetRType() override = 0;
    void Intersect(Ray &ray, ThreadInfo *threadinfo) override = 0;
    void Collision(ThreadInfo *threadinfo, Vector &min, Vector &max) override = 0;
    void Normal(Hit &hit, Vector &n, Vector &d) override = 0;
};

class Quadrangle final : public QuadObject {
    //16 :  16
public:
    Vector m_p[4];     //48 :  64
    int m_polyid; //4 :  68 bytes

#ifdef _DEBUG1
    Quadrangle(const Vector &p1, const Vector &p2, const Vector &p3, const Vector &p4, const float psmin, const float psmax, const float ptmin, const float ptmax, const int ppolyid) : QuadObject(psmin, psmax, ptmin, ptmax) {
        if (!IsFinite(p4.x)) {
            int a;
            a = 0;            
        }
        m_p[0] = p1;       
        m_p[1] = p2;       
        m_p[2] = p3;       
        m_p[3] = p4;       
        m_polyid = ppolyid;
    }
#else
    Quadrangle(const Vector &p1, const Vector &p2, const Vector &p3, const Vector &p4, const float psmin, const float psmax, const float ptmin, const float ptmax, const int ppolyid)
    : QuadObject(psmin, psmax, ptmin, ptmax), m_p{p1, p2, p3, p4}, m_polyid(ppolyid) {}
#endif
    void Bounds(Vector &min, Vector &max);
    float GetMin(int axis);
    float GetMax(int axis);

    UCHAR GetRType() override {
        return QUADRANGLE;
    }

    void Intersect(Ray &ray, ThreadInfo *threadinfo) override;
    void Collision(ThreadInfo *threadinfo, Vector &min, Vector &max) override;
    void Normal(Hit &hit, Vector &n, Vector &d) override;
};

class PropQuadrangle final : public QuadObject {
public:
    int m_vertexid[4]{};
    int m_normalid[4]{};
    bool m_isquad{false};

    PropQuadrangle() : QuadObject(PROPQUADRANGLE) {}

    void Bounds(const Vector *vertexlist, Vector &minv, Vector &maxv) const;

    UCHAR GetRType() override {
        return PROPQUADRANGLE;
    }

    void Intersect(Ray &ray, ThreadInfo *threadinfo) override;
    void Collision(ThreadInfo *threadinfo, Vector &min, Vector &max) override;
    void Normal(Hit &hit, Vector &n, Vector &d) override;
};

class Bone;
class Light;

class InfluenceShape : public RObject {
public:
    SobolSampler m_sampler;            //Quasi-Random stochastic 2D and 4D sampler
    Matrix34 m_worldmatrix,        //Transformation from object to world space
             m_worldinverse,       //Transformation fromworld to object space
             m_screenmatrix,       //Transformation from object to jittered screen space
             m_screeninverse,      //Transformation from jittered screen to object space
             m_steadyscreenmatrix, //Transformation from object to non-jittered screen space
             m_steadyscreeninverse;
    //Falloff calculations
    RGBFloat m_radiance, //Radiance on light emitting surface for total solidangle
             m_exitance; //Exitance on light surface
    float m_k0{0.f}, m_k2{0.f}, m_k1{0.f};
    //Radiosity and SSS
    float m_solidangle{0.f};    //Total solid angle from which photons are emited
    int m_ID{0};            //ID for probabilistic selection and density estimations
    int m_emitedphotons{0}; //Number of photons emited from this light
    Bone *m_bone{nullptr};   //Base bone

    void SetMatrices();

    InfluenceShape() = default;

    [[nodiscard]] float GetAttenuation(float d) const {
        if (d > 1.f)
            return hash_math::rcp(m_k0 + m_k1 * d + m_k2 * d * d);
        d = 6.f - 5.f * d;
        return -0.2f / (m_k0 + m_k1 * d + m_k2 * d * d) + 1.2f;
    }

    void SetAttenuation(const float attenuation) {
        m_k0 = attenuation < 0.5f ? 1.f - attenuation * 2.f : 0.f;
        m_k2 = attenuation > 0.5f ? attenuation * 2.f - 1.f : 0.f;
        m_k1 = 1.f - MAX(m_k0, m_k2);
    }

    void ResetSampler() {
        m_sampler.Init();
        ++m_emitedphotons;
    }

    UCHAR GetRType() override = 0;
    virtual float Falloff(Vector) = 0;
    virtual Vector Attract(Vector, float) = 0;
    virtual Vector Vortex(Vector, Vector, float) = 0;
    virtual Vector Friction(Vector, Vector, float) = 0;
    virtual float Lighting(const Vector &, Vector &, float &) = 0;

    virtual float LightingNoCap(const Vector &) {
        return 1.f;
    }

    virtual float DistanceSqr() {
        return 1.f;
    }

    virtual void Transform() = 0;
    virtual bool IsInside(Vector p) = 0;
    virtual bool IsInsideWideSoftness(Vector p) = 0;
    virtual void FindBounds(Vector &min, Vector &max) = 0;
    virtual float InitRadiosity(Light *light, int id) = 0;
    virtual int SampleRadiance(RGBFloat &c, Vector &p, Vector &d) = 0;
    virtual int SampleIrradiance(RGBFloat &c, Vector &pl, Vector &p) = 0;
};

class RSphere final : public InfluenceShape {
    //Implemented in Ray\Sphere.cpp
public:
    Vector m_center;                 //Position of center of sphere is screen space
    float m_radius{0.f},               //Radius from width given property
          m_radiussqr{0.f},            //m_radius^2
          m_falloffdistance{0.f},      //Fall-Off distance, a given property
          m_falloffdistancesqr{0.f},   //m_falloffdistance^2
          m_softfalloffdistance{0.f},  //For volumetric, a given property
          m_falloffdistancedelta{0.f}; //m_falloffdistance - m_softfalloffdistance

    RSphere(float radius, float falloffdistance, float softfalloffdistance);

    void Init(float radius, float falloffdistance, float softfalloffdistance);

    bool Intersect(const Ray &ray, Vector &in, Vector &out) override;
    void Intersect(Ray &ray, ThreadInfo *threadinfo) override;
    void Normal(Hit &hit, Vector &n, Vector &d) override;
    float Falloff(Vector p) override;
    Vector Attract(Vector p, float magnitude) override;
    Vector Vortex(Vector oldp, Vector vel, float magnitude) override;
    Vector Friction(Vector p, Vector force, float magnitude) override;
    float Lighting(const Vector &p, Vector &l, float &distance) override;

    float DistanceSqr() override {
        return m_falloffdistancesqr;
    }

    void Transform() override;
    bool IsInside(Vector p) override;

    bool IsInsideWideSoftness(const Vector p) override {
        return IsInside(p);
    }

    void FindBounds(Vector &min, Vector &max) override;
    float InitRadiosity(Light *light, int id) override;
    int SampleRadiance(RGBFloat &c, Vector &p, Vector &d) override;
    int SampleIrradiance(RGBFloat &c, Vector &pl, Vector &p) override;

    UCHAR GetRType() override {
        return RSPHERE;
    }
};

class Cone final : public InfluenceShape {
    //Implemented in Ray\Cone.cpp
public:
    Vector m_base,                  //Position of the tip of the cone in screen space
           m_axis,                  //normalized bone direction in screen space
           m_capbase;               //Position of the light bone (the face of the cone) in screen space
    float m_angle{0.f},                //The cone angle in radian from the given properties.
          m_halfangle{0.f},            //m_angle/2;
          m_tanangle{0.f},             //hash_math::tan( m_angle ); Length of opposite side when side == 1.
          m_tananglesqr{0.f},          //m_tanangle^2;
          m_costotalangle{0.f},        //hash_math::cos( m_angle );
          m_softness{0.f},             //angle of the softness border from the given percentage property
          m_cossoftangle{0.f},         //hash_math::cos( m_softness );
          m_deltaangle{0.f},           //m_cossoftangle - m_costotalangle;
          m_radius{0.f},               //width/2 where width is a given property
          m_endradius{0.f},            //hash_math::sqrt((m_length/m_costotalangle)^2 - m_lengthsqr);
          m_length{0.f},               //The Fall-Off Distance given property
          m_lengthsqr{0.f},            //m_length^2;
          m_falloffdistance{0.f},      //m_length * (1-falloffsoftness) + m_radius / m_tanangle; where falloffsoftness is a volumetric given property
          m_falloffdistancedelta{0.f}, //m_length - ( m_length * (1-falloffsoftness) );
          m_cap{0.f};                  //moveback distance. The distance from face of the cone (pivot) back to the tip of the cone
    bool m_flat{false};             //Used to decide how to sample for radiosity

    Cone(float, float, float, float, float);

    void Init(float, float, float, float, float);

    bool Intersect(const Ray &ray, Vector &in, Vector &out) override;
    void Intersect(Ray &ray, ThreadInfo *threadinfo) override;
    void Normal(Hit &hit, Vector &n, Vector &d) override;
    float Falloff(Vector) override;
    Vector Attract(Vector, float) override;
    Vector Vortex(Vector, Vector, float) override;
    Vector Friction(Vector, Vector, float) override;
    float Lighting(const Vector &, Vector &, float &) override;
    float LightingNoCap(const Vector &p) override;

    float DistanceSqr() override {
        return m_lengthsqr;
    }

    void Transform() override;
    void Transform(const Matrix34 &screenmatrix);
    //bool     IsVisible(Vector, const Matrix34 &, float, float, float(*GetVolumetricZ)(int, int));
    bool IsInside(Vector p) override;

    bool IsInsideWideSoftness(const Vector p) override {
        return IsInside(p);
    }

    void FindBounds(Vector &min, Vector &max) override;
    float InitRadiosity(Light *light, int id) override;
    int SampleRadiance(RGBFloat &c, Vector &p, Vector &d) override;
    int SampleIrradiance(RGBFloat &c, Vector &pl, Vector &p) override;

    UCHAR GetRType() override {
        return RCONE;
    }
};

class Cylinder final : public InfluenceShape {
    //Implemented in Ray\Cylinder.cpp
public:
    Vector m_base,                        //Position of the tip of the cone in screen space
           m_axis;                        //normalized bone direction in screen space
    float m_radius{0.f},                     //Radius of the cylinder from Width, a given property
          m_radiussqr{0.f},                  //m_radius^2
          m_length{0.f},                     //Fall-Off distance, a given property
          m_lengthsqr{0.f},                  //m_length^2
          m_radiusfalloffdistance{0.f},      //Portion of outher ring of cylinder where softness falloff occurs, a given property
          m_softfalloffdistance{0.f},        //m_length * (1-falloffsoftness) + m_radius / m_tanangle; where falloffsoftness is a volumetric given property
          m_falloffdistancedelta{0.f},       //m_length - ( m_length * (1-falloffsoftness) );
          m_radiusfalloffdistancedelta{0.f}; //m_radius - m_radiusfalloffdistance

    Cylinder(float, float, float, float);

    void Init(float, float, float, float);

    bool Intersect(const Ray &ray, Vector &in, Vector &out) override;
    void Intersect(Ray &ray, ThreadInfo *threadinfo) override;
    void Normal(Hit &hit, Vector &n, Vector &d) override;
    float Falloff(Vector) override;
    Vector Attract(Vector, float) override;
    Vector Vortex(Vector, Vector, float) override;
    Vector Friction(Vector, Vector, float) override;
    float Lighting(const Vector &, Vector &, float &) override;

    float DistanceSqr() override {
        return m_lengthsqr;
    }

    void Transform() override;
    bool IsInside(Vector p) override;
    bool IsInsideWideSoftness(Vector p) override;
    void FindBounds(Vector &min, Vector &max) override;
    float InitRadiosity(Light *light, int id) override;
    int SampleRadiance(RGBFloat &c, Vector &p, Vector &d) override;
    int SampleIrradiance(RGBFloat &c, Vector &pl, Vector &p) override;

    UCHAR GetRType() override {
        return RCYLINDER;
    }
};

class Slab final : public InfluenceShape {
    //Implemented in Ray\Slab.cpp
public:
    Vector m_center,          //Position of the
           m_axis;            //normalized bone direction in screen space
    float m_length{0.f},         //Fall-Off distance, a given property
          m_softlength{0.f},     //m_length * (1-m_falloffsoftness); where falloffsoftness is a volumetric given property
          m_halfwidth{0.f},      //half width from the width given property
          m_softhalfwidth{0.f},  //portion of half width that falls-off on the borders of the box
          m_lengthdelta{0.f},    //m_length - m_softlength;
          m_halfwidthdelta{0.f}; //m_halfwidth - m_softhalfwidth;

    Slab(float, float, float, float);

    void Init(float, float, float, float);

    bool Intersect(const Ray &ray, Vector &in, Vector &out) override;
    void Intersect(Ray &ray, ThreadInfo *threadinfo) override;
    void Normal(Hit &hit, Vector &n, Vector &d) override;
    float Falloff(Vector) override;
    Vector Attract(Vector, float) override;
    Vector Vortex(Vector, Vector, float) override;
    Vector Friction(Vector, Vector, float) override;
    float Lighting(const Vector &, Vector &, float &) override;
    //void     ComputeSides(Vector &, float, float, float);
    float DistanceSqr() override {
        return 1.f;
    }

    void Transform() override;
    bool IsInside(Vector p) override;
    bool IsInsideWideSoftness(Vector p) override;
    void FindBounds(Vector &min, Vector &max) override;
    float InitRadiosity(Light *light, int id) override;
    int SampleRadiance(RGBFloat &c, Vector &p, Vector &d) override;
    int SampleIrradiance(RGBFloat &c, Vector &pl, Vector &p) override;

    UCHAR GetRType() override {
        return RBOX;
    }
};

bool IntersectEndCap(const Ray &ray, const Vector &base, float radius, const Vector &n, float &t, Vector &hitp);
